"Python3 program using common n-grams to highlight formulaic sequences in html."

##  FLICSHOW :
##
##  Python3 program to highlight formulaic sequences :
##  Copyright 2015, R.S. Forsyth.
##  first version : 20/02/2015
##  last revision : 30/05/2016

ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  paracopy, parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  bastats, ranx
from quaylib import  phi4, philite
##from quaylib import  corsums, corrval
from quaylib import  fivenum, midvalue
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars
from quaylib import  subtuple


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","flicshow"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfanum initial
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency (??used)
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("maxiglen",5,2,9),  ##  maximum # of tokens in an n-gram
    ("miniglen",2,1,4),  ##  minimum n-gram size 2use in extending n-grams
    ("topgrams",100,8,4096),  ##  number of most frequent n-grams to keep
    ("dropsubs",0,0,1),  ##  whether to drop shorter subgrams on o/p (??used)
    ("huemode",0,0,1),  ##  colour-code coverage by longest (0) or total (1).
    ##  ("metafile",''),  ##  no training corpus, just uses gram data
    ("metaflic",''),  ##  test metafile: docs 2b shown as html files
    ("flipfile",''),  ##  human-readable results file
    ("gramdata",''),  ##  file for input of n-grams (from outgrams.py)
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"flicshow.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "1.4"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Gramtail = "_gram"
Fliptail = "_flip"
Flictail = "_flic"
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 64 * 1024  ##  will only seek tokens in front chunk of text file
##  Maxtoks not used !!
##  boilerplate for head & tail of html :
HTMLhead = """
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html>
  <head>
    <meta http-equiv="content-Type" content="text/html; charset=utf-8" />
<title>  ZonQ  </title>
</head>
<body>
<font color='#001122' >
"""
HTMLtail = """
</body>
</html>
"""
Zonq = "ZonQ"
Redfont = "<br/> </font> <font color='#F80000' size=4 >"
Normfont = "<br/> </font> <font color='#001122' size=3 >"
Huefont = "</font> <font color='######', size=4 >"
Huelist = ["#001122","#00FED5","#000FF0","#00DE0F","#FF8800","#FA0000","#990099"]
##  colour codes 0:6, black,cyan(grey?),blue,green,orange,red,purple.
##  Make Huefonts :
Huefonts = [Normfont]
for h in range(1,7):
    hf = Huefont.replace("######",Huelist[h])
    Huefonts.append(hf)
    ##  print(h,hf)
    

##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if paradat.miniglen > paradat.maxiglen:
        paradat.miniglen = max(1,paradat.maxiglen//2)
        print("miniglen adjusted to",paradat.miniglen)
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of main (training) corpus :
    metaname = getattr(paradat,"metaflic",'')
    if not metaname:
        print("Can't proceed without metaflic file!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
        if paradat.cats < 1:
            print("Must have at least 1 category !!")
            print("Program halts.")
            return  -2
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].totchars = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)

    for c in paradat.catlist:
        print(c,catinfo[c].docs)
        ##  showobj(catinfo[c])
        ##  catinfo doesn't have freqtabs @present.

    ##  attach extra attributes to doclist :
    allfrex = {} ; allsnips = {}
    paradat.totsnips = addsnips(paradat,doclist)  ##  snippets redundant ??
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        add2dic(allfrex,d.freqtab)
        for snip in d.snipsets:  ##  also tabulate snippet frequency
            for v in snip:
                sf = allsnips.get(v,0)
                allsnips[v] = sf + 1
        ##  redundant ??
        d.text = ' '.join(d.toklist) + ' '
        d.textlen = len(d.text)
        catinfo[d.outcome].totchars += d.textlen  ##  total character count
    ##  print("snippets =",paradat.totsnips)  ##  ??
    paradat.doclist = doclist  ##  copy reference
    paradat.allfrex = allfrex  ##  ditto
    paradat.vocsize = len(allfrex)

    ##  prepare o/p file :
    if paradat.flipfile:
        listfile = paradat.flipfile
    else:
        listfile = paradat.outpath + paradat.jobname + Fliptail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    paradat.fo = open(paradat.outfile,'w',encoding="utf-8")

    parashow(paradat,paradat.fo)
    ##  parashow(paradat,paradat.listfile)

    ##  read in n-gram data :
    if os.path.isfile(paradat.gramdata):
        gramname = paradat.gramdata
    else:
        gramname = paradat.outpath + paradat.jobname + Gramtail + ".dat"
    print("N-grams to be read from",os.path.realpath(gramname),':')
    paradat.gramfile = open(gramname,'r',encoding="utf-8")
    ng = getgrams(paradat,catinfo)
    print(ng,"n-grams read from",os.path.realpath(paradat.gramfile.name))
    print("reference category : ",paradat.refcode)

    ##  write header info :
    print(paradat.dateline,file=paradat.listfile)
    print("parafile:",paradat.id,file=paradat.listfile)
    print("metaflic:",paradat.metaflic,file=paradat.listfile)
    print("miniglen:",paradat.miniglen,file=paradat.listfile)
    print("maxiglen:",paradat.maxiglen,file=paradat.listfile)
    print("topgrams:",paradat.topgrams,file=paradat.listfile)
    print(paradat.docs,paradat.cats,file=paradat.listfile)
    print('',file=paradat.listfile)

    ##  identify sequences covered in each category :
    for cn,catname in enumerate(paradat.catlist):
        for dc,d in enumerate(doclist):
            if catname == d.outcome:
                d.marx,d.topmarx = grammarx(paradat,catinfo,catname,d)
                ##  sanity check :
                if dc in [-1]:
                    print(d.name)
                    for j,t in enumerate(d.toklist):
                        if d.marx[j]:
                            print(j+1,t,d.marx[j])

    ##  compute coverage by cats & by docs :
    for catname in paradat.catlist:
        catinfo[catname].tokhits = 0
        catinfo[catname].charhits = 0
    sortlist = []
    for cn,catname in enumerate(paradat.catlist):
        for dc,d in enumerate(doclist):
            if catname == d.outcome:
                d.charhits,d.tokhits = covered(paradat,d)
                d.charate = d.charhits / (d.textlen+1)
                d.tokrate = d.tokhits / d.size
                ##  print(dc+1,d.name,d.charate,d.tokrate)
                catinfo[catname].charhits += d.charhits
                catinfo[catname].tokhits += d.tokhits
                sortlist.append((d.charate,d.tokrate,dc))
                
    ##  coverage by category :
    print("Category coverage% (characters, tokens) by frequent n-grams :\n",
          file=paradat.listfile)
    for cn,catname in enumerate(paradat.catlist):
        if not (catname in paradat.gramcats):
            continue  ##  skip unseen class code
        cc = catinfo[catname].charhits / catinfo[catname].totchars
        ##  ??  + .docs
        tc = catinfo[catname].tokhits / catinfo[catname].tottoks
        print(cn,catname.ljust(16),round(cc*100,4),round(tc*100,4),
              file=paradat.listfile)
    paradat.listfile.write(NL)

    ##  also mark extraneous categories as if from .refcode :
    for dc,d in enumerate(doclist):
        if d.outcome in paradat.gramcats:
            pass
        else:
            d.marx,d.topmarx = grammarx(paradat,catinfo,paradat.refcode,d)
            
    ##  create html files :
    prepath = paradat.outpath
    if not ("html" in prepath):  ##  best not2 clutter up outpath
        prepath = prepath + "html" + os.sep
    if not os.path.isdir(prepath):
        os.makedirs(prepath)
    for dc,d in enumerate(doclist):
        catname = str(d.outcome)
        root,fx = os.path.splitext(d.name)
        fn = prepath + root + Flictail + ".html"
        fo = open(fn,'w',encoding="utf-8")
        opening = HTMLhead.replace(Zonq,d.name+" : "+catname)
        fo.write(opening)
        last = 0 ; outcode = Normfont
        if paradat.huemode > 0:  ##  colour by summed coverage
            marxlist = d.marx
        else:  ##  colour by longest n-gram coverage
            marxlist = d.topmarx  ##  default
        for j,m in enumerate(marxlist):
            t = d.toklist[j]
            ##  decide which font required :
            if last != m:
                outcode = Huefonts[min(6,m)]  ##  font code
                ##  if m > last:
                if m>0 and last==0:
                    outcode = "<br/> " + outcode
                fo.write(outcode)
            ##  else no need to change font.
            fo.write(t+' ')
            last = m  ##  previous code
        fo.write(NL)
        fo.write(HTMLtail)
        fo.close()
    

    ##  finale :
    paradat.listfile.close()
    paradat.fo.close()
    print("listing written onto file",os.path.abspath(paradat.listfile.name))
    print("files on",paradat.outpath,"with _flic.html suffix to be viewed with a browser.")    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0




##  LINEOUT :
def lineout (d,s):
    "makes printable line from file-info tuple."

    cc,tc,dc = s
    ##  oc = d.outcome[:12]
    f = "{0:6d} {1:5d} {2:7.2f}{3:7.2f} "
    lout = f.format(d.textlen,d.size,100*cc,100*tc)
    ##  outcome & file name appended outside.

    return  lout


##  EXTRACT :
def extract (paradat,catinfo):
    "extracts & tabulates subsequences in d.marx."

    for c in paradat.catlist:
        catinfo[c].formfrex = {}

    for d in paradat.doclist:
        catname = d.outcome
        d.slabfrex = subseqs(d)
        add2dic(catinfo[catname].formfrex,d.slabfrex)

    return


##  SUBSEQS :
def subseqs (d,consmin=2):
    "returns frequency table of subseqs in d (skipping singletons by default)."

    slabtab = {}
    nt = len(d.toklist)
    d.marx.append(0)  ##  sentinel
    slab = []
    for j,f in enumerate(d.marx):
        if f > 0:  ##  start or continuation of sequence
            slab.append(d.toklist[j])  ##  append corresponding token
            ##  if slab == ["leaflet", "carefully"]:
                ##  print(j,d.name,d.toklist[j-2:j+7])
        else:  ##  sequence ends
            if len(slab) >= consmin:  ##  worth saving
                stup = tuple(slab)
                sf = slabtab.get(stup,0)
                slabtab[stup] = sf + 1
                slab = []  ##  re-start
            else:  ##  too short to save
                slab = []

    del d.marx[-1]  ##  remove sentinel
    return  slabtab


##  COVERED :
def covered (paradat,d):
    "finds coverage by n-grams in 1 document."

    n = len(d.marx)
    assert n == d.size
    ch = th = 0
    for p in range(n):
        if d.marx[p] > 0:  ##  covered by frequent n-grams
            th += 1
            ch += (len(d.toklist[p])+1)  ##  always count (trailing) space

    return  ch,th


##  GRAMMARX :
def grammarx (paradat,catinfo,catname,d):
    "marks locations in d covered by n-grams of type catname."

    ntox = len(d.toklist)
    marx = [0] * ntox
    topmarx = [0] * ntox
    if catname != d.outcome:
        print(d.name,d.outcome,"treated as from class",catname)
    ##  assert catname == d.outcome
    for item in catinfo[catname].gramlist:
        g = list(item[-1])  ##  n-gram itself, forced 2b a list
        glen = len(g)  ##  can't be zero
        if glen < paradat.miniglen:
            continue  ##  too short (also check too long?)
        locs = findlocs(d.toklist,g[0])  ##  find locs of initial token
        for p in locs:
            slab = d.toklist[p:p+glen]
            if slab == g:
                for j in range(p,p+glen):
                    marx[j] += 1  ##  hits at this token position
                    topmarx[j] = max(topmarx[j],glen)
                    ##  length of longest covering n-gram.

    return  marx,topmarx  ##  2b held with d


##  FINDLOCS :
def findlocs (toklist,item):
    "returns locations of item in token list."

    locs = []
    p = 0
    nt = len(toklist)
    while p < nt:
        try:
            p = toklist.index(item,p)
            locs.append(p)
            p += 1
        except ValueError as xx:
            break

    return  locs


##  GETGRAMS :
def getgrams (paradat,catinfo):
    "gets n-grams from gramfile into catinfo sublist."

    datlines = paradat.gramfile.readlines()
    nl = len(datlines)
    print(nl)
    catname = "zonQ"
    ng = 0
    cmax = 0
    paradat.gramcats = set()
    for c in paradat.catlist:
        catinfo[c].gramlist = []  ##  initial gramlist
    for j,datline in enumerate(datlines):
        if not datline:
            continue  ##  skip blanks
        if datline[0] == '#':  ##  category header
            bits = datline.split()
            catname = bits[1]
            catsize = numerize(bits[-1])
            if not (catname in paradat.catlist):
                print("Additional class name :",catname)
                catinfo[catname] = Sack(catname)  ##  prepare 4new category
                catinfo[catname].gramlist = []
            paradat.gramcats.add(catname)
            if catsize > cmax:
                cmax = catsize
                refcode = catname
        elif " (" in datline:
            p = datline.find(" (")
            slab = datline[p:].strip()
            g = eval(slab)  ##  should become a tuple
            if len(g[-1]) >= paradat.miniglen:  ##  last item is n-gram
                ng += 1
                catinfo[catname].gramlist.append(g)  ##  still has size & freq
            ##  else too short 2b worth using
        else:
            pass
    ##  also reset reference category :
    paradat.refcode = refcode
    
    return  ng  ##  .gramcats also altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
