"Python3 program to make a minimal metafile from a given folder."

##  MINIMETA :
##
##  Python3 metafile maker :
##  Copyright 2011-2013, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 25/11/2011
##  last revision : 06/11/2013


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
import  time, random

    
##  local libraries :
from globdefs import *
from glib import numeric, numerize
from glib import openfile, paraget, parashow, showobj
from glib import termfrex, tokline
from glib import metaget, metachex
from glib import toklines
from glib import sibpath, parafind
##  from statlib import bastats, geomean
##  from statlib import ranx, midvalue


##  parameter info :
Parapath = os.path.expanduser('~')+os.sep+"parapath"  ##  will be added to parameter file search list
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("corpath",''),  ##  input folder
    ("metazero",''),  ##  o/p metafile
    ("filextn",''),
    ("targval","00",),  ##  default target value (-> doctype column)
    ("targname","doctype",),  ##  name of target column (third column)
    ("jobname","minimeta"),  ##  name to link related files
    ("wordonly",1,0,1),  ##  redundant ?!
    ("casefold",1,0,1),  ##  ditto ?!
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"minimeta.txt"),  ##  listing file
    ##  logfile ?
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  Global vars :
Version = "4.2"
Dumptail = "_meta"
Feffmarx = ["\ufeff","\uffef"]  ##  to detect Windows faffing around with feff


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()
    
    parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.path.sep
    if not paradat.corpath.endswith(os.path.sep):
        paradat.corpath = paradat.corpath + os.path.sep
    ##  paradat.filepath = paradat.corpath
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    ##  paradat.punxtab = punktran()  ##  uses default equivalents

    ##  open listing file :
    paradat.outname = paradat.outfile[:]
    paradat.outfile = open(paradat.outfile,'w',encoding="utf8")

    ##  prepare o/p zero-level metafile :
    if paradat.metazero:
        paradat.dumpname = paradat.metazero
    else:  ##  make a default name
        paradat.dumpname = paradat.outpath + paradat.jobname + Dumptail + ".txt"
    ##  make sure folder exists :
    dn = os.path.dirname(paradat.dumpname)
    print(dn)
    if not os.path.isdir(dn):
        os.makedirs(dn)
    paradat.dumpfile = open(paradat.dumpname,'w',encoding="utf-8")
    
    parashow(paradat,paradat.outfile)

    filelist = os.listdir(paradat.corpath)
    nf = len(filelist)
    filelist.sort()

    fc = 0 ; entries = 0
    heading = "prepath\tfilename\t" + paradat.targname + NL
    paradat.dumpfile.write(heading)
    for fn in filelist:
        selected = paradat.filextn == '' or fn.lower().endswith(paradat.filextn.lower())
        ##  print(fc,fn,selected)
        fullpath = paradat.corpath + fn
        if selected and os.path.exists(fullpath):
            catcode = paradat.targval  ##  default target value (4all docs)
            if paradat.targval == '_':  ##  secret sign !
                upos = fn.find('_')
                if upos > 0:
                    catcode = fn[:upos]  ##  extract prefix as targval
            lout = paradat.corpath + Tabchar + fn + Tabchar + catcode
            print(lout,file=paradat.dumpfile)
            entries += 1
        else:  ##  something wrong ?
            print(fc,entries,fn,"omitted from metafile listing.")
        fc += 1

    paradat.dumpfile.close()
    paradat.outfile.close()
    
    ##  finale :
    print(fc,"files read.") ; print(entries,"items written.")
    print("\nOutput listing on :",paradat.outfile.name)
    print("Results dumped onto:",paradat.dumpname,NL)
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,7),"seconds.")
    ##  shows secs since kickoff.
    
    return  0




##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname



            
##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
