"Python3 program to list frequent n-grams."

##  OUTGRAMS :
##
##  Python3 program to find & dump frequent (token) n-grams :
##  Copyright 2015, R.S. Forsyth.
##  first version : 09/02/2015
##  last revision : 30/05/2016

ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  bastats, ranx
from quaylib import  phi4, philite
##from quaylib import  corsums, corrval
from quaylib import  fivenum, midvalue
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars
from quaylib import  subtuple


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","outgrams"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfanum initial
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("maxiglen",5,2,9),  ##  maximum # of tokens in an n-gram
    ("miniglen",2,1,4),  ##  minimum # of tokens in an n-gram
    ("maxtops",20,0,144),  ##  maximum # of frequent stop-words
    ("topgrams",100,8,4096),  ##  number of most frequent n-grams to keep
    ("dropsubs",0,0,1),  ##  whether to drop shorter subgrams on o/p
    ("metafile",''),  ##  training corpus
    ("trainmet",''),  ##  ditto (takes precedence)
    ("gramdata",''),  ##  machine-readable results file
    ("gramlist",''),  ##  human-readable results file
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"outgrams.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "1.3"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Dumptail = "_gram"
Termtail = "_list"
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 64 * 1024  ##  will only seek tokens in front chunk of text file


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if paradat.trainmet:  ##  preferred case, trainmet takes precedence
        paradat.metafile = paradat.trainmet
        print(paradat.metafile,"to be used to indicate training texts.")
    elif paradat.metafile:
        print(paradat.metafile,"to be used to indicate training texts.")
    ##  else error, caught just below.
    if paradat.miniglen > paradat.maxiglen:
        paradat.miniglen = max(1,paradat.maxiglen//2)
        print("miniglen adjusted to",paradat.miniglen)
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of main (training) corpus :
    metaname = getattr(paradat,"metafile",'')
    if not metaname:
        print("Can't proceed without trainmet/metafile!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
        if paradat.cats < 2:
            print("Must have at least 2 categories !!")
            print("Program halts.")  ##  strictly necessary ??
            return  -2
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)
    print("reference category : ",paradat.refcode)

    for c in paradat.catlist:
        print(c,catinfo[c].docs)
        ##  showobj(catinfo[c])
        ##  catinfo doesn't have freqtabs @present.
        catinfo[c].charsize = 0

    ##  attach extra attributes to doclist :
    allfrex = {} ; allsnips = {}
    paradat.totsnips = addsnips(paradat,doclist)
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        add2dic(allfrex,d.freqtab)
        for snip in d.snipsets:  ##  also tabulate snippet frequency
            for v in snip:
                sf = allsnips.get(v,0)
                allsnips[v] = sf + 1
        ##  need to add into catinfo ??
        d.text = ' '.join(d.toklist) + ' '
        d.textlen = len(d.text)
        catinfo[d.outcome].charsize += d.textlen  ##  total character count
    print("snippets =",paradat.totsnips)  ##  used to define stop-words
    paradat.doclist = doclist  ##  copy reference
    paradat.allfrex = allfrex  ##  ditto
    paradat.vocsize = len(allfrex)
    for j in []:  ##  sanity check
        d = doclist[j]
        print(d.dnum,d.name,d.outcome,d.size,d.vocsize,d.snippets)
        for j in [0,-1]:
            s = d.snipsets[j]
            print(len(s),sorted(s))

    ##  prepare o/p files :
    if not os.path.isdir(paradat.outpath):
        os.makedirs(paradat.outpath)  ##  needs to be created
    if paradat.gramdata:
        dumpname = paradat.gramdata
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    ##  also need separate results file :
    if paradat.gramlist:
        listfile = paradat.gramlist
    else:
        listfile = paradat.outpath + paradat.jobname + Termtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    paradat.fo = open(paradat.outfile,'w',encoding="utf-8")

    parashow(paradat,paradat.fo)
    ##  parashow(paradat,paradat.listfile)

    ##  write header info :
    print(paradat.dateline,file=paradat.listfile)
    print("parafile:",paradat.id,file=paradat.listfile)
    print("metafile:",paradat.metafile,file=paradat.listfile)
    print("miniglen:",paradat.miniglen,file=paradat.listfile)
    print("maxiglen:",paradat.maxiglen,file=paradat.listfile)
    print("topgrams:",paradat.topgrams,file=paradat.listfile)
    print("dropsubs:",paradat.dropsubs,file=paradat.listfile)
    print("maxtops :",paradat.maxtops,file=paradat.listfile)
    print(paradat.docs,paradat.cats,file=paradat.listfile)
    print('',file=paradat.listfile)

    ##  first find frequent words :
    toplist = topfrex(allsnips,paradat.maxtops)
    paradat.stopset = set()
    for j,t in enumerate(toplist):
        if j == 0:  ##  write heading
            print("Stop-word listing :",file=paradat.listfile)
        f,w = t
        print(j+1,f,w,round(100*f/paradat.totsnips,2),
              file=paradat.listfile)
        paradat.stopset.add(w)
    paradat.listfile.write(NL)

    print("n-grams by category :\n",
            file=paradat.listfile)
    ##  next find frequent n-grams :
    for cn,catname in enumerate(paradat.catlist):
        gramtab = gramseek(paradat,catinfo,cn,catname)
        print(catname,catinfo[catname].tottoks,"tokens",
              file=paradat.listfile)
        print('#',catname,catinfo[catname].tottoks,
              catinfo[catname].charsize,file=paradat.dumpfile)
        ##  order first by freq then length :
        toplist = freqlen(paradat,gramtab)
        ##  for j,t in enumerate(toplist[:26]):
            ##  print(j,t)
        totchars = 0 ; ttox = 0
        for j,t in enumerate(toplist[:paradat.topgrams]):
            freq = t[0]
            gram = t[-1]
            ch = ' '.join(gram)
            totchars += len(ch)  ##  blank gaps included
            ttox += len(gram)  ##  tokens only
            print(j+1,freq,len(gram),len(ch),ch,file=paradat.listfile)
        print("mean toksize =",round(ttox/paradat.topgrams,4),
              file=paradat.listfile)
        print("mean charsize =",round(totchars/paradat.topgrams,4),NL,
              file=paradat.listfile)
        ##  also by length then freq :
        toplist = lenfreq(paradat,gramtab)
        ##  for j,t in enumerate(toplist[:paradat.topgrams]):
            ##  print(j+1,t,file=paradat.dumpfile)
        gout(paradat,catname,catinfo,toplist)

    ##  finale :
    paradat.dumpfile.close()
    paradat.listfile.close()
    paradat.fo.close()
    print("gram listing written on file",os.path.abspath(paradat.listfile.name))
    print("gram dump written onto file",os.path.abspath(paradat.dumpfile.name))    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0




##  GOUT :
def gout (paradat,catname,catinfo,toplist):
    "puts out sorted toplist of n-grams by size."

    written = []  ##  grams already written
    for n in range(paradat.maxiglen,paradat.miniglen-1,-1):
        ##  step down by # of tokens.
        counter = 0
        for t in toplist:
            size,nc,freq,g = t
            if size != n:
                continue
            if paradat.dropsubs > 0 and already(g,written):
                continue  ##  already written, dropsubs decides whether to omit
            counter += 1
            print(counter,t,file=paradat.dumpfile)
            written.append(g)  ##  permits dropping of shorter subsequences
            if counter >= paradat.topgrams:
                break  ## enough done at this size
            
    return


##  ALREADY :
def already (g,written):
    "has tuple g already been written?"

    for t in written:
        if subtuple(g,t):
            return  True

    return  False  ##  gets here if not found


##  LENFREQ :
def lenfreq (paradat,gtab):
    "orders gtab by length then frequency."

    sortlist = []
    for g in gtab:
        f = gtab[g]
        gc = ' '.join(g)
        if f > 2:  ##  hapaxes & dislegomena dropped
            sortlist.append((len(g),f,len(gc),g))  ##  charsize as tiebreak

    sortlist.sort(reverse=True)
    return  sortlist


##  FREQLEN :
def freqlen (paradat,gtab):
    "orders gtab by frequency then length."

    sortlist = []
    for g in gtab:
        f = gtab[g]
        nc = len(' '.join(g))
        if f > 2:  ##  hapaxes & dislegomena dropped
            sortlist.append((f,len(g),nc,g))

    sortlist.sort(reverse=True)
    return  sortlist


##  GRAMSEEK :
def gramseek (paradat,catinfo,cn,catname):
    "seeks n-grams in a given category."

    gramtab = {}
    dc = 0
    for d in paradat.doclist:
        if d.outcome != catname:
            continue  ##  skip other classes of document
        dc += 1
        for i in range(d.size):
            slab = d.toklist[i:i+paradat.maxiglen]
            ##  print(i,slab)
            for j in range(paradat.miniglen,len(slab)+1):
                ##  print(slab[0:j])
                gram = tuple(slab[0:j])
                if set(gram) <= paradat.stopset:
                    pass  ##  consists entirely of stop-words
                elif gram in gramtab:
                    gramtab[gram] += 1
                else:
                    gramtab[gram] = 1

        if dc == 1:
            print(catname)

    return  gramtab


##  TOPFREX :
def topfrex (vtab,maxtops):
    "returns high-freq items from vtab dictionary in descending order."

    sortlist = [(vtab[k],k) for k in vtab]
    sortlist.sort(reverse=True)

    return  sortlist[:maxtops]


##  NATS :
def nats (paradat,toktuple):
    "sums info in nats of tokens in toktuple."

    nsum = 0.0
    dnom = paradat.tottoks + paradat.vocsize
    for t in toktuple:
        p = (paradat.allfrex.get(t,0)+1) / dnom
        nsum -= math.log(p)

    return  nsum


##  ALFS :
def alfs (s,p=1):
    "number of alphabetics in a string or sequence."

    a = 0
    if type(s) == type('s'):
        for ch in s:
            if ch.isalpha():
                a += 1
    else:  ##  sequence
        for item in s:
            a += alfs(item)**p

    return  a
        



##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
