"Python3 program using common n-grams to estimate text attribution."

##  TAVERNS :
##
##  Textual Affinity Values Employing Repeated N-gram Sequences :
##  Copyright 2015, R.S. Forsyth.
##  first version : 23/02/2015
##  last revision : 30/05/2016

ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  paracopy, parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  bastats, ranx
from quaylib import  phi4, philite
##from quaylib import  corsums, corrval
from quaylib import  fivenum, midvalue
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars
from quaylib import  subtuple
from quaylib import  sckappa2


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","taverns"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfanum initial
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency (??used)
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("maxiglen",5,2,9),  ##  maximum # of tokens in an n-gram
    ("miniglen",2,1,4),  ##  minimum n-gram size 2use in extending n-grams
    ("topgrams",100,8,4096),  ##  number of most frequent n-grams to keep
    ("dropsubs",0,0,1),  ##  whether to drop shorter subgrams on o/p (??used)
    ("outforms",100,10,256),  ##  number of top forms to write out
    ("contrast",0,0,1),  ##  whether to use contrastive coverage
    ##  ("trainmet",''),  ##  uses gram data, not metafile or trainmet
    ("testmeta",''),  ##  test metafile: docs 2b compared to formulexicon
    ("blocfile",''),  ##  machine-readable results file
    ("formfile",''),  ##  human-readable results file
    ("gramdata",''),  ##  file for input of n-grams (from outgrams.py)
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"taverns.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "2.2"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Gramtail = "_gram"
Formtail = "_ales"
Bloctail = "_blox"
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 64 * 1024  ##  will only seek tokens in front chunk of text file
Nullcat = "ZonQ$$!!"  ##  null category name


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if paradat.miniglen > paradat.maxiglen:
        paradat.miniglen = max(1,paradat.maxiglen//2)
        print("miniglen adjusted to",paradat.miniglen)
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of testing corpus :
    metaname = getattr(paradat,"testmeta",'')
    if not metaname:  ##  go straight to test corpus ??
        print("Can't proceed without testmeta!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    ##  these are test docs, probably different from training data :
    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
        if paradat.cats < 1:
            print("Must have at least 1 category !!")
            print("Program halts.")
            return  -2
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].totchars = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  class with most tokens
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)

    dmax = catinfo[paradat.refcode].docs
    for c in paradat.catlist:
        cd = catinfo[c].docs
        print(c,cd)
        if cd > dmax:
            dmax = cd
            paradat.refcode = c
        ##  catinfo doesn't have freqtabs @present.
    print("provisional reference category :",paradat.refcode)
    
    ##  attach extra attributes to doclist :
    allfrex = {} ; allsnips = {}
    paradat.totsnips = addsnips(paradat,doclist)
    ##  snippets useless ?!
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        add2dic(allfrex,d.freqtab)
        for snip in d.snipsets:  ##  also tabulate snippet frequency
            for v in snip:
                sf = allsnips.get(v,0)
                allsnips[v] = sf + 1
        ##  need to add into catinfo ??
        d.text = ' '.join(d.toklist) + ' '
        d.textlen = len(d.text)
        catinfo[d.outcome].totchars += d.textlen  ##  total character count
    print("snippets =",paradat.totsnips)
    paradat.doclist = doclist  ##  copy reference
    paradat.allfrex = allfrex  ##  ditto
    paradat.vocsize = len(allfrex)

    ##  prepare o/p files :
    if paradat.blocfile:
        blocname = paradat.blocfile
    else:
        blocname = paradat.outpath + paradat.jobname + Bloctail + ".dat"
    paradat.dumpfile = open(blocname,'w',encoding="utf8")
    ##  also need separate results file :
    if paradat.formfile:
        listfile = paradat.formfile
    else:
        listfile = paradat.outpath + paradat.jobname + Formtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    paradat.fo = open(paradat.outfile,'w',encoding="utf-8")

    ##  read in n-gram data :
    if os.path.isfile(paradat.gramdata):
        gramname = paradat.gramdata
    else:
        gramname = paradat.outpath + paradat.jobname + Gramtail + ".dat"
    print("N-grams to be read from",os.path.realpath(gramname),':')
    paradat.gramfile = open(gramname,'r',encoding="utf-8")
    ng = getgrams(paradat,catinfo)
    ##  catinfo[*].gramlist created.
    print(ng,"n-grams read from",os.path.realpath(paradat.gramfile.name))
    nc = 0
    paradat.gramtags = []  ##  cats in n-gram file
    for c in catinfo:
        if catinfo[c].gramlist:
            nc += 1
            paradat.gramtags.append(c)
    print(nc,"n-gram categories.")
    paradat.gramtags.sort()
    paradat.gramcats = nc
    ##  print("actual reference category :",paradat.refcode)
    print("gram cats",paradat.gramtags)
    print("test cats",paradat.catlist)

    parashow(paradat,paradat.fo)

    ##  write header info :
    print(paradat.dateline,file=paradat.listfile)
    print("parafile:",paradat.id,file=paradat.listfile)
    print("gramfile:",gramname,file=paradat.listfile)
    print("testmeta:",paradat.testmeta,file=paradat.listfile)
    print("miniglen:",paradat.miniglen,file=paradat.listfile)
    print("maxiglen:",paradat.maxiglen,file=paradat.listfile)
    print("topgrams:",paradat.topgrams,file=paradat.listfile)
    print("contrast:",paradat.contrast,file=paradat.listfile)
    print(paradat.docs,paradat.cats,file=paradat.listfile)
    print('',file=paradat.listfile)

    ##  identify sequences covered in each category :
    for cn,catname in enumerate(paradat.catlist):
        if not catinfo[catname].gramlist:
            continue  ##  absent from n-gram file
        for dc,d in enumerate(doclist):
            if catname == d.outcome:
                d.marx = grammarx(paradat,catinfo,catname,d)
                ##  sanity check :
                if dc in [-1]:
                    print(d.name)
                    for j,t in enumerate(d.toklist):
                        if d.marx[j]:
                            print(j+1,t,d.marx[j])

    ##  extract & sort subsequences :
    extract(paradat,catinfo)
    for c,catname in enumerate(paradat.catlist):
        sortlist = []
        ndox = catinfo[catname].docs
        for t in catinfo[catname].formfrex:
            f = catinfo[catname].formfrex[t]
            nt = len(t)  ##  size in tokens
            st = ' '.join(t)
            x = (len(st)+1) * f  / catinfo[catname].totchars
            ##  character coverage.
            sortlist.append((round(100*x,4),f,nt,len(st),t))
        sortlist.sort(reverse=True)
        print(c,catname,ndox,catinfo[catname].tottoks,catinfo[catname].totchars,
              file=paradat.dumpfile)
        jmax = min(len(sortlist),paradat.outforms)  ##  sometimes very few
        for j in range(jmax):
            print(sortlist[j],file=paradat.dumpfile)

    ##  rank docs against each gram-base :
    if os.path.isfile(paradat.testmeta):
        ##  compute coverage on test docs :
        if paradat.contrast <= 0:
            cont = ''
        else:
            cont = "contrastive "
        for cn,catname in enumerate(paradat.gramtags):
            if not catinfo[catname].gramlist:
                continue  ##  class absent from gram data
            print("\nRanking by "+cont+"coverage of sequences from",catname,
                  file=paradat.listfile)
            print(catname)
            if paradat.contrast:  ##  prepare anti-gramlist :
                catinfo[Nullcat] = Sack(Nullcat)
                catinfo[Nullcat].gramlist = antigram(paradat,paradat,
                                                     catinfo,catname)
                paradat.fo.write("anti-"+catname+NL)
                for item in catinfo[Nullcat].gramlist:
                    print(item,file=paradat.fo)  ##  mainly 4checking
                paradat.fo.write(NL)
                sortlist = docranx(paradat,paradat,catinfo,catname,
                                   catinfo)  ##  contrastive marking
            else:  ##  positive marking only
                sortlist = docrank(paradat,paradat,catinfo,catname)
            sortlist.sort(reverse=True)
            ranktot = {}
            for j,s in enumerate(sortlist):
                cc,tc,dc = s
                d = doclist[dc]  ##  refers to testmeta files
                oc = str(d.outcome)
                if not hasattr(d,"catranx"):
                    d.catranx = {}
                ##  er = catinfo[oc].docs / 2 + 0.5  ##  midrank
                d.catranx[catname] = cc
                if oc in ranktot:
                    ranktot[oc] += j
                else:
                    ranktot[oc] = j
                lout = lineout(d,s)  ##  tuple to printable line
                print(str(j+1).rjust(4)+' '+lout+' '+oc[:12].ljust(12)+' '+d.name,
                      file=paradat.listfile)
            paradat.listfile.write(NL)
            for catname in paradat.catlist:
                ar = ranktot[catname] / catinfo[catname].docs
                print(catname,"mean rank =",round(ar+1,2),file=paradat.listfile)
        print(file=paradat.listfile)
        ##  predictive mode :
        ctab = {} ; seen = 0
        hitz = 0 ; nonnullz = 0
        predlist = []
        for d in paradat.doclist:  ##  find highest coverage
            mr = 0.0 ; rtot = 0.0
            pred = paradat.refcode
            if pred in d.catranx:
                mr = d.catranx[pred]
            for c in d.catranx:
                rr = d.catranx[c]  ##  rating not ranking !
                rtot += rr
                if rr > mr:  ##  must exceed
                    mr = rr ; pred = c
            ##  print(d.name,d.outcome,mr,pred)
            rr = mr / (rtot + Tiny)  ##  relative rating of predicted class
            predlist.append((rr,mr,d.name,d.outcome,pred))
            if d.outcome in paradat.gramtags:
                seen += 1
                if (mr > Tiny):  ##  nonzero rating
                    nonnullz += 1
                    if d.outcome == pred:
                        hitz += 1
            pair = (pred,d.outcome)
            f = ctab.get(pair,0)
            ctab[pair] = f+1
        ##  show classification results :
        predlist.sort(reverse=True)  ##  descending order
        predshow(paradat,predlist)
        outshow(paradat,ctab,seen,paradat.listfile)
        ##  append nonnull success rate :
        print("\nNumber of non-null instances =",nonnullz,file=paradat.listfile)
        print("Correct decisions in such cases =",hitz,file=paradat.listfile)
        pc = 100*hitz / max(1,nonnullz)
        print("Percent of such cases correct =",round(pc,2),
              file=paradat.listfile)

    ##  finale :
    paradat.dumpfile.close()
    paradat.listfile.close()
    paradat.fo.close()
    print("listing written onto file",os.path.abspath(paradat.listfile.name))
    print("blocdump written onto file",os.path.abspath(paradat.dumpfile.name))    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0




##  PREDSHOW :
def predshow (paradat,predlist):
    "shows ranked predictions."

    print("\nResults in classification mode :",file=paradat.listfile)
    h1 = "\n      relative   actual       categories"
    h2 = "rank  coverage% coverage%     pred : true      docname"
    fm = "{0:4d}  {1:8.2f} {2:8.2f}   {3:>8s} ~ {4:8s}  {5:s}"
    print(h1,file=paradat.listfile)
    print(h2,file=paradat.listfile)
    for j,p in enumerate(predlist):  ##  already sorted
        rr,r,dn,oc,pred = p  ##  unpack tuple
        lout = fm.format(j+1,100*rr,100*r,pred[:8].rjust(8),oc[:8],dn)
        if pred == oc:  ##  must be strings
            tick = '+'
        elif not (oc in paradat.gramtags):
            tick = '?'
        else:
            tick = '-'
        lout = lout.replace('~',tick,1)
        print(lout,file=paradat.listfile)

    print(file=paradat.listfile)
    return


##  OUTSHOW :
def outshow (paradat,cmat,seen,fo):
    "shows confusion matrix."
    
    ##  show confusion matrix :
    ptab,rtab = precall(cmat)
    keylist = sorted(cmat.keys())  ##  pairs
    wide = max([len(str(k)) for k in keylist])
    wide = wide // 2
    rc = paradat.docs
    ##  matshow & precall expect pred,true order :
    matshow(fo,cmat)
    sk = sckappa2(cmat,paradat.catlist)  ##  keylist ??
    k2 = round(sk[0],4)
    ##  print(keylist,paradat.catlist)
    hits = sk[2]
    print("Kappa value = ",k2,file=fo)
    print("Precision (%) by category :",file=fo)
    for c in sorted(ptab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(ptab[c],4),file=fo)
    print("Recall (%) by category :",file=fo)
    for c in sorted(rtab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(rtab[c],4),file=fo)

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below only apply to",seen,word+']',file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        
    return


##  PRECALL :
def precall (cmat):
    "returns precisions & recalls of confusion matrix held as freqtab of pairs."

    c1 = set() ; c2 = set()
    pairlist = sorted(cmat.keys())  ##  order in pair: predcat,truecat
    for k1,k2 in pairlist:
        c1.add(k1) ; c2.add(k2)
    c1 = sorted(c1) ; c2 = sorted(c2)
    ptab = {} ; rtab = {}
    ##  obtain totals :
    for p in pairlist:
        f = cmat[p]
        pn = ptab.get(p[0],0)
        ptab[p[0]] = pn + f
        rn = rtab.get(p[1],0)
        rtab[p[1]] = rn + f
    ##  print(ptab,rtab)
    ##  compute percentages :
    for c in c1:
        ok = cmat.get((c,c),0)
        ptab[c] = 100.0 * ok / ptab.get(c,1)  ##  precision in class c
    for c in c2:
        ok = cmat.get((c,c),0)
        rtab[c] = 100.0 * ok / rtab.get(c,1)  ##  recall of class c

    return  (ptab,rtab)


##  MATSHOW :
def matshow (fo,cmat,longest=Longname):
    "write confusion matrix (dict of pairs) to already open file."

    fo.write(NL + "Confusion matrix :" + NL + NL)
    ##  make local category list (could include unknown cats in holdout mode) :
    cats = set()
    predcats = set()
    for k1,k2 in cmat.keys():
        cats.add(k1) ; cats.add(k2)
        predcats.add(k1)
    catlist = sorted(cats)
    ##  find widest class name :
    w = 4
    for c in catlist:
        wx = len(str(c))
        if wx > w:
            w = wx
    w = min(longest,w)
    lout = "Truecat =".ljust(10+w)
    for c in catlist:
        lout = lout + str(c)[:longest].rjust(w+1)
    fo.write(lout + NL)
    for pred in sorted(predcats):
        lout = "Predcat : " + str(pred).ljust(w)
        for truecat in catlist:
            pair = (pred,truecat)
            f = cmat.get(pair,0)
            lout = lout + " {0:{1}d}".format(f,w)
        fo.write(lout + NL)

    fo.write(NL)
    return


##  ANTIGRAM :
def antigram (paradat1,paradat,catinfo,catname):
    "makes aggregate gramlist from other cats by rank."
    ##  paradat1 to go eventually.

    outlist = []
    nc = paradat.gramcats - 1  ##  number of other categories
    ##  print(catname,nc,paradat.gramcats-1)
    ##  what if nc < 1 ?!
    for size in range(paradat.maxiglen,paradat.miniglen-1,-1):
        gramtab = {}
        for cn,thiscat in enumerate(paradat.catlist):
            if catname == thiscat:
                continue  ##  only other classes
            rank = 0
            for g in catinfo[thiscat].gramlist:
                gg = g[-1]  ##  actual n-gram
                if len(gg) == size:
                    rank += 1 ; toprank = rank
                    if gg in gramtab:
                        gramtab[gg].append(rank)
                    else:
                        gramtab[gg] = [rank]
        ##  calculate average ranking :
        sortlist = []
        for gg in gramtab:
            rx = gramtab[gg]
            ng = len(rx)
            gs = ' '.join(gg)
            t = sum(rx)
            if ng < nc:
                t += ((nc-ng) * (1+toprank))
            meanrank = t / nc
            sortlist.append((meanrank,len(gs),gg))
        sortlist.sort()
        gmax = min(len(sortlist),paradat.topgrams)
        for j in range(gmax):
            outlist.append(sortlist[j])

    return  outlist


##  DOCRANX :
def docranx(paradat1,paradat,catinfo1,catname,catinfo):
    "computes differential coverage of catname seqs in paradat.doclist."

    sortlist = []
    for dc,d in enumerate(paradat.doclist):
        d.marx = grammarx(paradat,catinfo1,catname,d)  ##  positive marks
        ##  anti-marx :
        antimarx = grammarx(paradat,catinfo,Nullcat,d)  ##  negative marks
        for j in range(len(d.marx)):
            if antimarx[j] > 0:
                d.marx[j] = 0  ##  suppress if covered by alien grams
        formchar,formtoks = covered(paradat,d)
        cc = round(formchar/d.textlen,7)
        tc = round(formtoks/d.size,7)
        sortlist.append((cc,tc,dc))

    return  sortlist


##  DOCRANK :
def docrank(paradat1,paradat,catinfo1,catname):
    "computes coverage of catname seqs in paradat.doclist."

    sortlist = []
    for dc,d in enumerate(paradat.doclist):
        d.marx = grammarx(paradat,catinfo1,catname,d)
        formchar,formtoks = covered(paradat,d)
        cc = round(formchar/d.textlen,7)
        tc = round(formtoks/d.size,7)
        sortlist.append((cc,tc,dc))

    return  sortlist


##  LINEOUT :
def lineout (d,s):
    "makes printable line from file-info tuple."

    cc,tc,dc = s
    ##  oc = d.outcome[:12]
    f = "{0:6d} {1:5d} {2:7.2f}{3:7.2f} "
    lout = f.format(d.textlen,d.size,100*cc,100*tc)
    ##  outcome & file name appended outside.

    return  lout


##  EXTRACT :
def extract (paradat,catinfo):
    "extracts & tabulates subsequences in d.marx."

    for c in paradat.catlist:
        catinfo[c].formfrex = {}

    for d in paradat.doclist:
        if hasattr(d,"marx"):
            catname = d.outcome
            d.slabfrex = subseqs(d,paradat.miniglen)
            add2dic(catinfo[catname].formfrex,d.slabfrex)

    return


##  SUBSEQS :
def subseqs (d,consmin=2):
    "returns frequency table of subseqs in d (skipping singletons by default)."

    slabtab = {}
    nt = len(d.toklist)
    d.marx.append(0)  ##  sentinel
    slab = []
    for j,f in enumerate(d.marx):
        if f > 0:  ##  start or continuation of sequence
            slab.append(d.toklist[j])  ##  append corresponding token
            ##  if slab == ["leaflet", "carefully"]:
                ##  print(j,d.name,d.toklist[j-2:j+7])
        else:  ##  sequence ends
            if len(slab) >= consmin:  ##  worth saving
                stup = tuple(slab)
                sf = slabtab.get(stup,0)
                slabtab[stup] = sf + 1
                slab = []  ##  re-start
            else:  ##  too short to save
                slab = []

    del d.marx[-1]  ##  remove sentinel
    return  slabtab


##  COVERED :
def covered (paradat,d):
    "finds coverage by n-grams in 1 document."

    n = len(d.marx)
    assert n == d.size
    ch = th = 0
    for p in range(n):
        if d.marx[p] > 0:  ##  covered by frequent n-grams
            th += 1
            ch += (len(d.toklist[p])+1)  ##  always count (trailing) space

    return  ch,th


##  GRAMMARX :
def grammarx (paradat,catinfo,catname,d):
    "marks locations in d covered by n-grams of type catname."

    ntox = len(d.toklist)
    marx = [0] * ntox
    ##  catname doesn't have to be same as d.outcome
    for item in catinfo[catname].gramlist:
        g = list(item[-1])  ##  n-gram itself, forced 2b a list
        glen = len(g)  ##  can't be zero
        if glen < paradat.miniglen:
            continue  ##  too short
        locs = findlocs(d.toklist,g[0])  ##  find locs of initial token
        for p in locs:
            slab = d.toklist[p:p+glen]
            if slab == g:
                for j in range(p,p+glen):
                    marx[j] += 1  ##  hits at this token position

    return  marx  ##  2b held with d


##  FINDLOCS :
def findlocs (toklist,item):
    "returns locations of item in token list."

    locs = []
    p = 0
    nt = len(toklist)
    while p < nt:
        try:
            p = toklist.index(item,p)
            locs.append(p)
            p += 1
        except ValueError as xx:
            break

    return  locs


##  GETGRAMS :
def getgrams (paradat,catinfo):
    "gets n-grams from gramfile into catinfo sublist."

    datlines = paradat.gramfile.readlines()
    nl = len(datlines)
    print(nl)
    catname = "zonQ"
    ng = 0
    ttox = 0
    for c in paradat.catlist:
        catinfo[c].gramlist = []  ##  initial gramlist
    for j,datline in enumerate(datlines):
        if not datline:
            continue  ##  skip blanks
        if datline[0] == '#':  ##  category header
            bits = datline.split()
            catname = bits[1]
            if not (catname in paradat.catlist):
                print("Unexpected class name :",catname)
                catinfo[catname] = Sack(catname)
                catinfo[catname].gramlist = []  ##  class absent from testmeta
            ntox = numerize(bits[2])
            catinfo[catname].toksize = ntox
            ##  bits[2] should be toksize.
            ##  keep track of largest class from gramlist file :
            ##  no longer (26/05/2016).
        elif " (" in datline:
            p = datline.find(" (")
            slab = datline[p:].strip()
            g = eval(slab)  ##  should become a tuple
            if len(g[-1]) >= paradat.miniglen:  ##  last item is n-gram
                ng += 1
                catinfo[catname].gramlist.append(g)  ##  still has size & freq
            ##  else too short 2b worth using
        else:
            pass

    return  ng


##  TOPFREX :
def topfrex (vtab,maxtops):
    "returns high-freq items from vtab dictionary in descending order."

    sortlist = [(vtab[k],k) for k in vtab]
    sortlist.sort(reverse=True)

    return  sortlist[:maxtops]


##  NATS :
def nats (paradat,toktuple):
    "sums info in nats of tokens in toktuple."

    nsum = 0.0
    dnom = paradat.tottoks + paradat.vocsize
    for t in toktuple:
        p = (paradat.allfrex.get(t,0)+1) / dnom
        nsum -= math.log(p)

    return  nsum


##  ALFS :
def alfs (s,p=1):
    "number of alphabetics in a string or sequence."

    a = 0
    if type(s) == type('s'):
        for ch in s:
            if ch.isalpha():
                a += 1
    else:  ##  sequence
        for item in s:
            a += alfs(item)**p

    return  a
        



##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
